@@ -3,6 +3,7 @@ use std::rc::Rc;
 
 use super::error::{PolarError, PolarResult};
 use super::formatting::{source_lines, ToPolarString};
+use super::partial::simplify_bindings;
 use super::sources::*;
 use super::terms::*;
 use super::traces::*;
@@ -54,6 +55,7 @@ enum Step {
     /// Step **in**. Will break on the next query.
     Into,
     Error,
+    Rule,
 }
 
 /// VM breakpoints.
@@ -69,6 +71,7 @@ pub enum DebugEvent {
     Query,
     Pop,
     Error(PolarError),
+    Rule,
 }
 
 /// Tracks internal debugger state.
@@ -80,6 +83,7 @@ pub struct Debugger {
     /// - `Some(step)`: View the stopping logic in
     ///   [`maybe_break`](struct.Debugger.html#method.maybe_break).
     step: Option<Step>,
+    last: Option<String>,
 }
 
 impl Debugger {
@@ -128,30 +132,24 @@ impl Debugger {
                     message: format!("{}\nERROR: {}\n", message, error.to_string()),
                 })
             }
+            (Step::Rule, DebugEvent::Rule) => self.break_query(vm),
             _ => None,
         })
     }
 
     pub fn break_msg(&self, vm: &PolarVirtualMachine) -> Option<String> {
-        vm.trace.last().and_then(|trace| {
-            if let Trace {
-                node: Node::Term(q),
-                ..
-            } = &**trace
-            {
-                match q.value() {
-                    Value::Expression(Operation {
-                        operator: Operator::And,
-                        args,
-                    }) if args.len() == 1 => None,
-                    _ => {
-                        let source = self.query_source(q, &vm.kb.read().unwrap().sources, 3);
-                        Some(format!("{}\n\n{}\n", vm.query_summary(q), source))
-                    }
+        vm.trace.last().and_then(|trace| match trace.node {
+            Node::Term(ref q) => match q.value() {
+                Value::Expression(Operation {
+                    operator: Operator::And,
+                    args,
+                }) if args.len() == 1 => None,
+                _ => {
+                    let source = self.query_source(q, &vm.kb.read().unwrap().sources, 3);
+                    Some(format!("{}\n\n{}\n", vm.query_summary(q), source))
                 }
-            } else {
-                None
-            }
+            },
+            Node::Rule(ref r) => Some(vm.rule_source(r)),
         })
     }
 
@@ -186,7 +184,13 @@ impl Debugger {
             }
         }
         let parts: Vec<&str> = command.split_whitespace().collect();
-        match *parts.get(0).unwrap_or(&"help") {
+        let default_command = match self.last.take() {
+            Some(s) => s,
+            _ => "help".to_owned(),
+        };
+        let command = *parts.get(0).unwrap_or(&&default_command[..]);
+        self.last = Some(String::from(command));
+        match command {
             "c" | "continue" | "q" | "quit" => self.step = None,
 
             "n" | "next" | "over" => {
@@ -204,6 +208,9 @@ impl Debugger {
             "e" | "error" => {
                 self.step = Some(Step::Error)
             }
+            "r" | "rule" => {
+                self.step = Some(Step::Rule)
+            }
             "l" | "line" => {
                 let lines = parts.get(1).and_then(|s| s.parse().ok()).unwrap_or(0);
                 return Some(Goal::Debug {
@@ -309,24 +316,7 @@ impl Debugger {
                     let vars: Vec<Binding> = parts[1..]
                         .iter()
                         .map(|name| {
-                            // *** variable name mapping ***
-                            // if the requested variable is bound, then we return that binding.
-                            // otherwise, we look for the matching bound temp variable with the
-                            // highest numeric component in its name, and return that binding
-                            // if we find it. otherwise, show that the variable is unbound.
-                            let var = Symbol::new(name);
-                            let bindings = vm.bindings(true);
-                            bindings.get(&var).cloned().map_or_else(|| {
-                                let prefix = KnowledgeBase::temp_prefix(name);
-                                bindings.keys()
-                                    .filter_map(|k| k.0.strip_prefix(&prefix).and_then(|i|
-                                        i.parse::<i64>().map_or(None, |i| Some((k, i)))))
-                                    .max_by(|a, b| a.1.cmp(&b.1))
-                                    .map_or_else(
-                                        || Binding(Symbol::new(name), Term::new_temporary(Value::Variable(Symbol::new("<unbound>")))),
-                                        |b| Binding(Symbol::new(format!("{}@{}", name, b.0.0).as_str()), bindings.get(b.0).unwrap().clone()))
-                            },
-                            |val| Binding(var, val))
+                            get_binding_for_var(name, vm)
                         })
                         .collect();
                     return Some(show(&vars));
@@ -353,6 +343,7 @@ impl Debugger {
   o[ut]                   Step out of the current query stack level to the next query in the level above.
   g[oal]                  Step to the next goal of the Polar VM.
   e[rror]                 Step to the next error.
+  r[ule]                  Step to the next rule.
   l[ine] [<n>]            Print the current line and <n> lines of context.
   query [<i>]             Print the current query or the query at level <i> in the query stack.
   stack | trace           Print the current query stack.
@@ -368,3 +359,35 @@ impl Debugger {
         None
     }
 }
+
+/// *** variable name mapping ***
+/// if the requested variable is bound, then we return that binding.
+/// otherwise, we look for the matching bound temp variable with the
+/// highest numeric component in its name, and return that binding
+/// if we find it. otherwise, show that the variable is unbound.
+pub fn get_binding_for_var(name: &str, vm: &PolarVirtualMachine) -> Binding {
+    let var = Symbol::new(name);
+    let bindings = simplify_bindings(vm.bindings(true)).unwrap();
+    bindings.get(&var).cloned().map_or_else(
+        || {
+            let prefix = KnowledgeBase::temp_prefix(name);
+            bindings
+                .keys()
+                .filter_map(|k| {
+                    k.0.strip_prefix(&prefix)
+                        .and_then(|i| i.parse::<i64>().map_or(None, |i| Some((k, i))))
+                })
+                .max_by(|a, b| a.1.cmp(&b.1))
+                .map_or_else(
+                    || Binding(sym!(name), Term::from(sym!("<unbound>"))),
+                    |b| {
+                        Binding(
+                            sym!(format!("{}@{}", name, b.0 .0).as_str()),
+                            bindings.get(b.0).unwrap().clone(),
+                        )
+                    },
+                )
+        },
+        |val| Binding(var, val),
+    )
+}
