@@ -41,6 +41,7 @@ pub struct ErrorContext {
     pub source: Source,
     pub row: usize,
     pub column: usize,
+    pub include_location: bool,
 }
 
 impl PolarError {
@@ -63,16 +64,22 @@ impl PolarError {
                         source: source.clone(),
                         row,
                         column,
+                        include_location: false,
                     });
                 }
                 _ => {}
             },
-            (_, Some(source), Some(term)) => {
+            (e, Some(source), Some(term)) => {
                 let (row, column) = crate::lexer::loc_to_pos(&source.src, term.offset());
                 self.context.replace(ErrorContext {
                     source: source.clone(),
                     row,
                     column,
+                    // @TODO(Sam): find a better way to include this info
+                    include_location: matches!(
+                        e,
+                        ErrorKind::Runtime(RuntimeError::UnhandledPartial { .. })
+                    ),
                 });
             }
             _ => {}
@@ -241,6 +248,12 @@ pub enum ParseError {
 
 impl fmt::Display for ErrorContext {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        // @TODO(Sam): find a better way to incorporate this info
+        if self.include_location {
+            writeln!(f, "found in:")?;
+            write!(f, "{}", self.source.src.split('\n').nth(self.row).unwrap())?;
+            write!(f, "\n{}^", " ".repeat(self.column))?;
+        }
         write!(f, " at line {}, column {}", self.row + 1, self.column + 1)?;
         if let Some(ref filename) = self.source.filename {
             write!(f, " in file {}", filename)?;
@@ -326,7 +339,7 @@ pub enum RuntimeError {
         sym: Symbol,
     },
     StackOverflow {
-        msg: String,
+        limit: usize,
     },
     QueryTimeout {
         msg: String,
@@ -341,6 +354,10 @@ pub enum RuntimeError {
     IncompatibleBindings {
         msg: String,
     },
+    UnhandledPartial {
+        var: Symbol,
+        term: Term,
+    },
 }
 
 impl RuntimeError {
@@ -367,7 +384,9 @@ impl fmt::Display for RuntimeError {
                 write!(f, "Type error: {}", msg)
             }
             Self::UnboundVariable { sym } => write!(f, "{} is an unbound variable", sym.0),
-            Self::StackOverflow { msg } => write!(f, "Hit a stack limit: {}", msg),
+            Self::StackOverflow { limit } => {
+                write!(f, "Goal stack overflow! MAX_GOALS = {}", limit)
+            }
             Self::QueryTimeout { msg } => write!(f, "Query timeout: {}", msg),
             Self::Application { msg, stack_trace } => {
                 if let Some(stack_trace) = stack_trace {
@@ -379,6 +398,26 @@ impl fmt::Display for RuntimeError {
             Self::IncompatibleBindings { msg } => {
                 write!(f, "Attempted binding was incompatible: {}", msg)
             }
+            Self::UnhandledPartial { var, term } => {
+                write!(
+                    f,
+                    "Found an unhandled partial in the query result: {var}
+
+This can happen when there is a variable used inside a rule
+which is not related to any of the query inputs.
+
+For example: f(_x) if y.a = 1 and y.b = 2;
+
+In this example, the variable `y` is constrained by `a = 1 and b = 2`,
+but we cannot resolve these constraints without further information.
+
+The unhandled partial is for variable {var}.
+The expression is: {expr}
+",
+                    var = var,
+                    expr = term.to_polar(),
+                )
+            }
         }
     }
 }
@@ -424,6 +463,7 @@ impl fmt::Display for ParameterError {
 pub enum ValidationError {
     InvalidRule { rule: String, msg: String },
     InvalidRuleType { rule_type: String, msg: String },
+    UndefinedRule { rule_name: String },
     // TODO(lm|gj): add ResourceBlock and SingletonVariable.
 }
 
@@ -436,6 +476,9 @@ impl fmt::Display for ValidationError {
             Self::InvalidRuleType { rule_type, msg } => {
                 write!(f, "Invalid rule type: {} {}", rule_type, msg)
             }
+            Self::UndefinedRule { rule_name } => {
+                write!(f, r#"Call to undefined rule "{}""#, rule_name)
+            }
         }
     }
 }
