@@ -2,9 +2,12 @@ use std::collections::{HashMap, HashSet};
 
 use crate::error::ParameterError;
 use crate::error::{PolarError, PolarResult};
+use crate::validations::check_undefined_rule_calls;
+use crate::visitor::{walk_term, Visitor};
 
 pub use super::bindings::Bindings;
 use super::counter::Counter;
+use super::diagnostic::Diagnostic;
 use super::resource_block::ResourceBlocks;
 use super::resource_block::{ACTOR_UNION_NAME, RESOURCE_UNION_NAME};
 use super::rules::*;
@@ -17,8 +20,8 @@ enum RuleParamMatch {
     False(String),
 }
 
+#[cfg(test)]
 impl RuleParamMatch {
-    #[cfg(test)]
     fn is_true(&self) -> bool {
         matches!(self, RuleParamMatch::True)
     }
@@ -32,10 +35,10 @@ pub struct KnowledgeBase {
     /// Map of class name -> MRO list where the MRO list is a list of class instance IDs
     mro: HashMap<Symbol, Vec<u64>>,
 
-    /// Map from loaded files to the source ID
-    pub loaded_files: HashMap<String, u64>,
-    /// Map from source code loaded to the filename it was loaded as
-    pub loaded_content: HashMap<String, String>,
+    /// Map from filename to source ID for files loaded into the KB.
+    loaded_files: HashMap<String, u64>,
+    /// Map from contents to filename for files loaded into the KB.
+    loaded_content: HashMap<String, String>,
 
     rules: HashMap<Symbol, GenericRule>,
     rule_types: RuleTypes,
@@ -107,8 +110,20 @@ impl KnowledgeBase {
         generic_rule.add_rule(Arc::new(rule));
     }
 
+    pub fn validate_rules(&self) -> Vec<Diagnostic> {
+        let mut diagnostics = self.validate_rule_calls();
+        if let Err(e) = self.validate_rule_types() {
+            diagnostics.push(Diagnostic::Error(e));
+        }
+        diagnostics
+    }
+
+    fn validate_rule_calls(&self) -> Vec<Diagnostic> {
+        check_undefined_rule_calls(self)
+    }
+
     /// Validate that all rules loaded into the knowledge base are valid based on rule types.
-    pub fn validate_rules(&self) -> PolarResult<()> {
+    fn validate_rule_types(&self) -> PolarResult<()> {
         for (rule_name, generic_rule) in &self.rules {
             if let Some(types) = self.rule_types.get(rule_name) {
                 // If a type with the same name exists, then the parameters must match for each rule
@@ -265,7 +280,16 @@ impl KnowledgeBase {
                             }
                         }
                         if !success {
-                            return Ok(RuleParamMatch::False(format!("Rule specializer {} on parameter {} must be a member of rule type specializer {}", rule_instance.tag,index, rule_type_instance.tag)));
+                            let mut err = format!("Rule specializer {} on parameter {} must be a member of rule type specializer {}", rule_instance.tag,index, rule_type_instance.tag);
+                            if rule_type_instance.tag == sym!("Actor") {
+                                err.push_str(&format!("
+
+\tPerhaps you meant to add an actor block to the top of your policy, like this:
+
+\t  actor {} {{}}", rule_instance.tag));
+                            }
+
+                            return Ok(RuleParamMatch::False(err));
                         }
                     }
                     if !self.param_fields_match(&rule_type_instance.fields, &rule_instance.fields) {
@@ -378,6 +402,19 @@ impl KnowledgeBase {
                     Value::Variable(_),
                     Some(Value::Pattern(rule_spec)),
                 ) => self.check_pattern_param(index, rule_spec, rule_type_spec)?,
+                // RuleType has an instance pattern specializer but rule has no specializer
+                (
+                    Value::Variable(_),
+                    Some(Value::Pattern(Pattern::Instance(InstanceLiteral { tag, .. }))),
+                    Value::Variable(parameter),
+                    None,
+                ) => RuleParamMatch::False(format!(
+                    "Parameter `{parameter}` expects a {tag} type constraint.
+
+\t{parameter}: {tag}",
+                    parameter = parameter,
+                    tag = tag
+                )),
                 // RuleType has specializer but rule doesn't
                 (Value::Variable(_), Some(rule_type_spec), Value::Variable(_), None) => {
                     RuleParamMatch::False(format!(
@@ -481,6 +518,8 @@ impl KnowledgeBase {
             })
             .collect::<PolarResult<Vec<RuleParamMatch>>>()
             .map(|results| {
+                // TODO(gj): all() is short-circuiting -- do we want to gather up *all* failure
+                // messages instead of just the first one?
                 results.iter().all(|r| {
                     if let RuleParamMatch::False(msg) = r {
                         failure_message = msg.to_owned();
@@ -565,55 +604,6 @@ impl KnowledgeBase {
         self.resource_blocks.clear();
     }
 
-    /// Removes a file from the knowledge base by finding the associated
-    /// `Source` and removing all rules for that source, and
-    /// removes the file from loaded files.
-    ///
-    /// Optionally return the source for the file, returning `None`
-    /// if the file was not in the loaded files.
-    pub fn remove_file(&mut self, filename: &str) -> Option<String> {
-        self.loaded_files
-            .get(filename)
-            .cloned()
-            .map(|src_id| self.remove_source(src_id))
-    }
-
-    /// Removes a source from the knowledge base by finding the associated
-    /// `Source` and removing all rules for that source. Will
-    /// also remove the loaded files if the source was loaded from a file.
-    pub fn remove_source(&mut self, source_id: u64) -> String {
-        // remove from rules
-        self.rules.retain(|_, gr| {
-            let to_remove: Vec<u64> = gr.rules.iter().filter_map(|(idx, rule)| {
-                matches!(rule.source_info, SourceInfo::Parser { src_id, ..} if src_id == source_id)
-                    .then(||*idx)
-            }).collect();
-
-            for idx in to_remove {
-                gr.remove_rule(idx);
-            }
-            !gr.rules.is_empty()
-        });
-
-        // remove from sources
-        let source = self
-            .sources
-            .remove_source(source_id)
-            .expect("source doesn't exist in KB");
-        let filename = source.filename;
-
-        // remove queries
-        self.inline_queries
-            .retain(|q| q.get_source_id() != Some(source_id));
-
-        // remove from files
-        if let Some(filename) = filename {
-            self.loaded_files.remove(&filename);
-            self.loaded_content.retain(|_, f| f != &filename);
-        }
-        source.src
-    }
-
     fn check_file(&self, src: &str, filename: &str) -> PolarResult<()> {
         match (
             self.loaded_content.get(src),
@@ -649,46 +639,67 @@ impl KnowledgeBase {
     }
 
     pub fn set_error_context(&self, term: &Term, error: impl Into<PolarError>) -> PolarError {
-        let source = term
-            .get_source_id()
-            .and_then(|id| self.sources.get_source(id));
+        /// `GetSource` will talk a term and return the _first_ piece of source
+        /// info it finds
+        struct GetSource<'kb> {
+            kb: &'kb KnowledgeBase,
+            source: Option<Source>,
+            term: Option<Term>,
+        }
+
+        impl<'kb> Visitor for GetSource<'kb> {
+            fn visit_term(&mut self, t: &Term) {
+                if self.source.is_none() {
+                    self.source = t
+                        .get_source_id()
+                        .and_then(|id| self.kb.sources.get_source(id));
+
+                    if self.source.is_none() {
+                        walk_term(self, t)
+                    } else {
+                        self.term = Some(t.clone())
+                    }
+                }
+            }
+        }
+
+        let mut source_getter = GetSource {
+            kb: self,
+            source: None,
+            term: None,
+        };
+        source_getter.visit_term(term);
+        let source = source_getter.source;
+        let term = source_getter.term;
         let error: PolarError = error.into();
-        error.set_context(source.as_ref(), Some(term))
+        error.set_context(source.as_ref(), term.as_ref())
     }
 
-    pub fn rewrite_shorthand_rules(&mut self) -> PolarResult<()> {
+    pub fn rewrite_shorthand_rules(&mut self) -> Vec<Diagnostic> {
         let mut errors = vec![];
 
         errors.append(
             &mut super::resource_block::check_all_relation_types_have_been_registered(self),
         );
 
-        // TODO(gj): Emit all errors instead of just the first.
-        if !errors.is_empty() {
-            return Err(errors[0].clone());
-        }
-
         let mut rules = vec![];
-        for (resource_block, shorthand_rules) in &self.resource_blocks.shorthand_rules {
+        for (resource_name, shorthand_rules) in &self.resource_blocks.shorthand_rules {
             for shorthand_rule in shorthand_rules {
-                match shorthand_rule.as_rule(resource_block, &self.resource_blocks) {
+                match shorthand_rule.as_rule(resource_name, &self.resource_blocks) {
                     Ok(rule) => rules.push(rule),
                     Err(error) => errors.push(error),
                 }
             }
         }
 
-        // TODO(gj): Emit all errors instead of just the first.
-        if !errors.is_empty() {
-            return Err(errors[0].clone());
+        if errors.is_empty() {
+            // Add the rewritten rules to the KB.
+            for rule in rules {
+                self.add_rule(rule);
+            }
         }
 
-        // Add the rewritten rules to the KB.
-        for rule in rules {
-            self.add_rule(rule);
-        }
-
-        Ok(())
+        errors.into_iter().map(Diagnostic::Error).collect()
     }
 
     pub fn is_union(&self, maybe_union: &Term) -> bool {
@@ -1266,11 +1277,11 @@ mod tests {
         kb.add_rule(rule!("f", ["x"; instance!(sym!("Fruit"))]));
 
         assert!(matches!(
-            kb.validate_rules().err().unwrap(),
-            PolarError {
+            kb.validate_rules().first().unwrap(),
+            Diagnostic::Error(PolarError {
                 kind: ErrorKind::Validation(ValidationError::InvalidRule { .. }),
                 ..
-            }
+            })
         ));
 
         // Rule type does not apply if it doesn't have the same name as a rule
@@ -1279,7 +1290,7 @@ mod tests {
         kb.add_rule(rule!("f", ["x"; instance!(sym!("Orange"))]));
         kb.add_rule(rule!("g", ["x"; instance!(sym!("Fruit"))]));
 
-        kb.validate_rules().unwrap();
+        assert!(kb.validate_rules().is_empty());
 
         // Rule type does apply if it has the same name as a rule even if different arity
         kb.clear_rules();
@@ -1287,11 +1298,11 @@ mod tests {
         kb.add_rule(rule!("f", ["x"; instance!(sym!("Orange"))]));
 
         assert!(matches!(
-            kb.validate_rules().err().unwrap(),
-            PolarError {
+            kb.validate_rules().first().unwrap(),
+            Diagnostic::Error(PolarError {
                 kind: ErrorKind::Validation(ValidationError::InvalidRule { .. }),
                 ..
-            }
+            })
         ));
         // Multiple templates can exist for the same name but only one needs to match
         kb.clear_rules();
