@@ -11,6 +11,7 @@ mod bindings;
 mod counter;
 pub mod data_filtering;
 mod debugger;
+pub mod diagnostic;
 pub mod error;
 pub mod events;
 mod folder;
@@ -30,6 +31,6 @@ mod runnable;
 pub mod sources;
 pub mod terms;
 pub mod traces;
+mod validations;
 mod visitor;
 mod vm;
-mod warnings;
