@@ -6,7 +6,6 @@ use permute::permute;
 
 use std::cell::RefCell;
 use std::collections::{BTreeMap, HashMap};
-use std::iter::FromIterator;
 
 use polar_core::{
     error::*,
@@ -749,7 +748,8 @@ fn test_two_rule_bodies_not_nested() -> TestResult {
     let mut p = Polar::new();
     p.load_str(
         r#"f(x) if a(x);
-           f(1);"#,
+           f(1);
+           a(_x) if false;"#,
     )?;
     qvar(&mut p, "f(x)", "x", values![1]);
     Ok(())
@@ -761,7 +761,8 @@ fn test_two_rule_bodies_nested() -> TestResult {
     p.load_str(
         r#"f(x) if a(x);
            f(1);
-           a(x) if g(x);"#,
+           a(x) if g(x);
+           g(_x) if false;"#,
     )?;
     qvar(&mut p, "f(x)", "x", values![1]);
     Ok(())
@@ -1576,6 +1577,32 @@ fn test_unknown_specializer_warning() -> TestResult {
     Ok(())
 }
 
+#[test]
+fn test_missing_actor_hint() -> TestResult {
+    let p = Polar::new();
+
+    p.register_constant(sym!("Organization"), term!(true))?;
+    p.register_constant(sym!("User"), term!(true))?;
+
+    let policy = r#"
+resource Organization {
+    roles = ["owner"];
+    permissions = ["read"];
+
+    "read" if "owner";
+}
+
+has_role(user: User, "owner", organization: Organization) if
+    organization.owner_id = user.id;
+"#;
+    let err = p.load_str(policy).expect_err("Expected validation error");
+    assert!(matches!(&err.kind, ErrorKind::Validation(_)));
+    assert!(format!("{}", err)
+        .contains("Perhaps you meant to add an actor block to the top of your policy, like this:"));
+
+    Ok(())
+}
+
 #[test]
 fn test_and_or_warning() -> TestResult {
     let p = Polar::new();
@@ -1600,21 +1627,29 @@ fn test_and_or_warning() -> TestResult {
 
     p.clear_rules();
     p.load_str("f(x) if x = 1 and x > 1 or x < 3;")?;
-    let msg = p.next_message().unwrap();
-    assert!(matches!(&msg.kind, MessageKind::Warning));
-    assert_eq!(
-        &msg.msg,
-        "Expression without parentheses could be ambiguous. \nPrior to 0.20, `x and y or z` would parse as `x and (y or z)`. \nAs of 0.20, it parses as `(x and y) or z`, matching other languages. \n\n\n001: f(x) if x = 1 and x > 1 or x < 3;\n             ^"
-    );
+    let mut messages = vec![];
+    while let Some(msg) = p.next_message() {
+        messages.push(msg);
+    }
+    assert!(messages.iter().any(|msg| {
+      matches!(&msg.kind, MessageKind::Warning) &&
+      (msg.msg ==
+        "Expression without parentheses could be ambiguous. \nPrior to 0.20, `x and y or z` would parse as `x and (y or z)`. \nAs of 0.20, it parses as `(x and y) or z`, matching other languages. \n\n\n001: f(x) if x = 1 and x > 1 or x < 3;\n             ^")
+    }));
 
     p.clear_rules();
     p.load_str("f(x) if x = 1 or x > 1 and x < 3;")?;
-    let msg = p.next_message().unwrap();
-    assert!(matches!(&msg.kind, MessageKind::Warning));
-    assert_eq!(
-        &msg.msg,
-        "Expression without parentheses could be ambiguous. \nPrior to 0.20, `x and y or z` would parse as `x and (y or z)`. \nAs of 0.20, it parses as `(x and y) or z`, matching other languages. \n\n\n001: f(x) if x = 1 or x > 1 and x < 3;\n                      ^"
-    );
+
+    let mut msgs: Vec<Message> = vec![];
+    while let Some(msg) = p.next_message() {
+        msgs.push(msg);
+    }
+    assert!(msgs.iter().any(|msg| {
+      matches!(&msg.kind, MessageKind::Warning) &&
+      (msg.msg ==
+        "Expression without parentheses could be ambiguous. \nPrior to 0.20, `x and y or z` would parse as `x and (y or z)`. \nAs of 0.20, it parses as `(x and y) or z`, matching other languages. \n\n\n001: f(x) if x = 1 or x > 1 and x < 3;\n                      ^")
+    }));
+
     Ok(())
 }
 
@@ -1623,12 +1658,15 @@ fn test_print() -> TestResult {
     // TODO: If POLAR_LOG is on this test will fail.
     let p = Polar::new();
     p.load_str("f(x,y,z) if print(x, y, z);")?;
+    let mut messages = vec![];
     let message_handler = |output: &Message| {
-        assert!(matches!(&output.kind, MessageKind::Print));
-        assert_eq!(&output.msg, "1, 2, 3");
+        messages.push(output.clone());
     };
     let q = p.new_query("f(1, 2, 3)", false)?;
     let _results = query_results!(q, @msgs message_handler);
+    assert!(messages
+        .iter()
+        .any(|msg| { matches!(&msg.kind, MessageKind::Print) && (msg.msg == "1, 2, 3") }));
     Ok(())
 }
 
@@ -1660,6 +1698,51 @@ fn test_partial_grounding() -> TestResult {
     Ok(())
 }
 
+#[test]
+fn test_dict_destructuring() -> TestResult {
+    let mut p = Polar::new();
+    p.load_str(
+        r#"
+        foo(x, _: {x});
+        goo(x, y) if y matches {x};
+        moo(x, {x});
+        roo(a, {a, b: a});
+        too(a, _: {a, b: a});
+   "#,
+    )?;
+    for s in &["foo", "goo"] {
+        qeval(&mut p, &format!("{}(1, {{x: 1}})", s));
+        qnull(&mut p, &format!("{}(2, {{x: 1}})", s));
+        qnull(&mut p, &format!("{}(1, {{x: 2}})", s));
+        qeval(&mut p, &format!("{}(2, {{x: 2, y: 3}})", s));
+    }
+
+    qeval(&mut p, "moo(1, {x: 1})");
+    qnull(&mut p, "moo(2, {x: 1})");
+    qnull(&mut p, "moo(1, {x: 2})");
+    qnull(&mut p, "moo(2, {x: 2, y: 3})");
+
+    qeval(&mut p, "roo(1, {a: 1, b: 1})");
+    qeval(&mut p, "too(1, {a: 1, b: 1})");
+    qnull(&mut p, "roo(1, {a: 1, b: 1, c: 2})");
+    qeval(&mut p, "too(1, {a: 1, b: 1, c: 2})");
+    Ok(())
+}
+
+#[ignore]
+#[test]
+fn test_dict_destructuring_broken() -> TestResult {
+    let mut p = Polar::new();
+    p.load_str(
+        r#"
+        too(a, _: {a, b: a});
+   "#,
+    )?;
+    // currently hits an unimplemented code path in vm.rs
+    qeval(&mut p, "a={a} and too(a, {a, b: a})");
+    Ok(())
+}
+
 #[test]
 fn test_rest_vars() -> TestResult {
     let mut p = Polar::new();
@@ -1696,7 +1779,7 @@ fn test_rest_vars() -> TestResult {
         "a = [1, *b] and b = [2, *c] and c=[3] and 1 in a and 2 in a and 3 in a",
     );
 
-    let a = &var(&mut p, "[*_] in [*a] and [*b] in [*_] and b = 1", "a")[0];
+    let a = &var(&mut p, "[*c] in [*a] and [*b] in [*d] and b = 1", "a")[0];
     // check that a isn't bound to [b]
     assert!(!matches!(a, Value::List(b) if matches!(b[0].value(), Value::Number(_))));
     Ok(())
@@ -1830,6 +1913,7 @@ fn test_matches() {
     qeval(&mut p, "x = {foo: 1, bar: 2} and x matches {foo: 1}");
     qnull(&mut p, "x = {foo: 1} and x matches {foo: 1, bar: 2}");
     qnull(&mut p, "x = {foo: 1} and x matches {foo: 2}");
+    qeval(&mut p, "x matches Integer and x = 1");
 }
 
 #[test]
@@ -2259,17 +2343,6 @@ fn test_list_matches() {
     );
 }
 
-#[test]
-#[allow(clippy::unnecessary_wraps)]
-fn error_on_binding_expressions_and_patterns_to_variables() -> TestResult {
-    qruntime!(
-        "x matches y",
-        RuntimeError::TypeError { msg: m, .. },
-        m == "cannot unify patterns directly `x` = `y{}`"
-    );
-    Ok(())
-}
-
 #[test]
 fn test_builtin_iterables() {
     let mut p = Polar::new();
@@ -2333,19 +2406,44 @@ fn test_lookup_in_rule_head() -> TestResult {
 #[test]
 fn test_default_rule_types() -> TestResult {
     let p = Polar::new();
+
     // This should fail
     let e = p
         .load_str(r#"has_permission("leina", "eat", "food");"#)
         .expect_err("Expected validation error");
     assert!(matches!(e.kind, ErrorKind::Validation(_)));
+    assert!(matches!(
+        p.next_message(),
+        Some(Message {
+            kind: MessageKind::Warning,
+            msg
+        }) if msg.starts_with("Your policy does not contain an allow rule")
+    ));
+
     let e = p
         .load_str(r#"has_role("leina", "eater", "food");"#)
         .expect_err("Expected validation error");
     assert!(matches!(e.kind, ErrorKind::Validation(_)));
+    assert!(matches!(
+        p.next_message(),
+        Some(Message {
+            kind: MessageKind::Warning,
+            msg
+        }) if msg.starts_with("Your policy does not contain an allow rule")
+    ));
+
     let e = p
         .load_str(r#"has_relation("leina", "eater", "food");"#)
         .expect_err("Expected validation error");
     assert!(matches!(e.kind, ErrorKind::Validation(_)));
+    assert!(matches!(
+        p.next_message(),
+        Some(Message {
+            kind: MessageKind::Warning,
+            msg
+        }) if msg.starts_with("Your policy does not contain an allow rule")
+    ));
+
     let e = p
         .load_str(r#"allow("leina", "food");"#)
         .expect_err("Expected validation error");
@@ -2381,3 +2479,51 @@ fn test_default_rule_types() -> TestResult {
     assert!(p.next_message().is_none());
     Ok(())
 }
+
+#[test]
+fn test_suggested_rule_specializer() -> TestResult {
+    let p = Polar::new();
+
+    let repo_instance = ExternalInstance {
+        instance_id: 1,
+        constructor: None,
+        repr: None,
+    };
+    let repo_term = term!(Value::ExternalInstance(repo_instance.clone()));
+    let repo_name = sym!("Repository");
+    p.register_constant(repo_name.clone(), repo_term)?;
+    p.register_mro(repo_name, vec![repo_instance.instance_id])?;
+
+    let user_instance = ExternalInstance {
+        instance_id: 2,
+        constructor: None,
+        repr: None,
+    };
+    let user_term = term!(Value::ExternalInstance(user_instance.clone()));
+    let user_name = sym!("User");
+    p.register_constant(user_name.clone(), user_term)?;
+    p.register_mro(user_name, vec![user_instance.instance_id])?;
+
+    let policy = r#"
+actor User {}
+resource Repository {
+    permissions = ["read"];
+    roles = ["contributor"];
+
+    "read" if "contributor";
+}
+
+has_role(actor: User, role_name, repository: Repository) if
+    role in actor.roles and
+    role_name = role.name and
+    repository = role.repository;
+"#;
+
+    let err = p.load_str(policy).expect_err("Expected validation error");
+    assert!(matches!(&err.kind, ErrorKind::Validation(_)));
+    assert!(format!("{}", err).contains(
+        "Failed to match because: Parameter `role_name` expects a String type constraint."
+    ));
+
+    Ok(())
+}
